@@ -143,8 +143,10 @@ static void exec_funcdef(const command_T *c, bool finally_exit)
     __attribute__((nonnull));
 
 static void exec_commands(command_T *c, exec_T type);
-static inline bool should_exit(const command_T *c)
-    __attribute__((nonnull));
+static bool is_errexit_condition(void)
+    __attribute__((pure));
+static inline bool is_errexit_condition_for(const command_T *c)
+    __attribute__((nonnull,pure));
 static inline void next_pipe(pipeinfo_T *pi, bool next)
     __attribute__((nonnull));
 static pid_t exec_process(
@@ -300,7 +302,7 @@ void exec_pipelines(const pipeline_T *p, bool finally_exit)
 	suppresserrexit |= p->pl_neg || p->next != NULL;
 
 	bool self = finally_exit && !doing_job_control_now
-	    && !p->next && !p->pl_neg && !any_trap_set;
+	    && !p->next && !p->pl_neg && !any_trap_set && !shopt_pipefail;
 	exec_commands(p->pl_commands, self ? E_SELF : E_NORMAL);
 	if (p->pl_neg) {
 	    if (laststatus == Exit_SUCCESS)
@@ -406,6 +408,8 @@ void exec_for(const command_T *c, bool finally_exit)
 	/* expand the words between "in" and "do" of the for command. */
 	if (!expand_line(c->c_forwords, &count, &words)) {
 	    laststatus = Exit_EXPERROR;
+	    if (is_errexit_condition())
+		exit_shell_with_status(laststatus);
 	    goto finish;
 	}
     } else {
@@ -537,6 +541,8 @@ done:
 
 fail:
     laststatus = Exit_EXPERROR;
+    if (is_errexit_condition())
+	exit_shell_with_status(laststatus);
     goto done;
 }
 
@@ -646,37 +652,21 @@ void exec_commands(command_T *c, exec_T type)
 
     handle_signals();
 
-    if (shopt_errexit && should_exit(c))
+    if (is_errexit_condition_for(c)) {
 	exit_shell_with_status(laststatus);
+    }
 
     comsfree(c);
 }
 
 /* Returns true if the shell should exit because of the `errexit' option. */
-bool should_exit(const command_T *c)
+bool is_errexit_condition(void)
 {
-    if (suppresserrexit)
+    if (!shopt_errexit || suppresserrexit)
 	return false;
     if (laststatus == Exit_SUCCESS)
 	return false;
 
-    /* If this is a multi-command pipeline, the commands are executed in
-     * subshells. Otherwise, we need to check the type of the command. */
-    if (c->next == NULL) {
-	switch (c->c_type) {
-	    case CT_SIMPLE:
-	    case CT_SUBSHELL:
-	    case CT_FUNCDEF:
-		break;
-	    case CT_GROUP:
-	    case CT_IF:
-	    case CT_FOR:
-	    case CT_WHILE:
-	    case CT_CASE:
-		return false;
-	}
-    }
-
 #if YASH_ENABLE_LINEEDIT
     if (le_state & LE_STATE_COMPLETING)
 	return false;
@@ -685,6 +675,33 @@ bool should_exit(const command_T *c)
     return true;
 }
 
+/* Returns true if the shell should exit because of the `errexit' option. */
+bool is_errexit_condition_for(const command_T *c)
+{
+    if (!is_errexit_condition())
+	return false;
+
+    /* If this is a multi-command pipeline, the commands are executed in
+     * subshells. Otherwise, we need to check the type of the command. */
+    if (c->next != NULL)
+	return true;
+
+    switch (c->c_type) {
+	case CT_SIMPLE:
+	case CT_SUBSHELL:
+	case CT_FUNCDEF:
+	    return true;
+	case CT_GROUP:
+	case CT_IF:
+	case CT_FOR:
+	case CT_WHILE:
+	case CT_CASE:
+	    return false;
+    }
+
+    assert(false);
+}
+
 /* Updates the contents of the `pipeinfo_T' to proceed to the next process
  * execution. `pi->pi_fromprevfd' and `pi->pi_tonextfds[PIPE_OUT]' are closed,
  * `pi->pi_tonextfds[PIPE_IN]' is moved to `pi->pi_fromprevfd', and,
@@ -798,6 +815,8 @@ pid_t exec_process(
     if (!open_redirections(c->c_redirs, &savefd)) {
 	/* On redirection error, the command is not executed. */
 	laststatus = Exit_REDIRERR;
+	if (is_errexit_condition())
+	    exit_shell_with_status(laststatus);
 	if (posixly_correct && !is_interactive_now && c->c_type == CT_SIMPLE &&
 		argc > 0 && is_special_builtin(argv0))
 	    finally_exit = true;
