@@ -1,6 +1,6 @@
 /* Yash: yet another shell */
 /* expand.c: word expansion */
-/* (C) 2007-2013 magicant */
+/* (C) 2007-2016 magicant */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -43,26 +43,37 @@
 #include "yash.h"
 
 
+/* characters that have special meanings in brace expansion, quote removal, and
+ * globbing. When an unquoted expansion includes these characters, they are
+ * backslashed to protect from unexpected side effects in succeeding expansion
+ * steps. */
+#define CHARS_ESCAPED L"\\\"\'{,}"
+
 static bool expand_and_split_words(
 	const wordunit_T *restrict w, plist_T *restrict list)
     __attribute__((nonnull(2)));
 
 /* data passed between expansion functions */
-struct expand_word_T {
+struct expand_four_T {
     plist_T *valuelist, *splitlist;
     xwcsbuf_T valuebuf;
     xstrbuf_T splitbuf;
-    bool putempty;
+    bool zeroword;
 };
+/* When "$@" appears during expansion and there is no positional parameter, the
+ * `zeroword' flag is set so that the quoted empty word can be removed later. */
 
-static bool expand_word(
-	const wordunit_T *restrict w, tildetype_T tilde, bool quoted,
-	plist_T *restrict valuelist, plist_T *restrict splitlist)
-    __attribute__((nonnull(4)));
-static bool expand_word_inner(const wordunit_T *restrict w,
-	tildetype_T tilde, bool quoted, bool rec,
-	struct expand_word_T *restrict e)
+static bool expand_four_and_remove_quotes(
+	const wordunit_T *restrict w,
+	tildetype_T tilde, bool processquotes, bool escapeall,
+	plist_T *restrict valuelist)
     __attribute__((nonnull(5)));
+static bool expand_four(const wordunit_T *restrict w,
+	tildetype_T tilde, bool processquotes, bool escapeall, bool rec,
+	struct expand_four_T *restrict e)
+    __attribute__((nonnull(6)));
+static void fill_splitbuf(struct expand_four_T *e, bool splittable)
+    __attribute__((nonnull));
 
 static wchar_t *expand_tilde(const wchar_t **ss,
 	bool hasnextwordunit, tildetype_T tt)
@@ -71,10 +82,8 @@ static wchar_t *expand_tilde(const wchar_t **ss,
 enum indextype_T { IDX_NONE, IDX_ALL, IDX_CONCAT, IDX_NUMBER, };
 
 static bool expand_param(const paramexp_T *restrict p, bool indq,
-	struct expand_word_T *restrict e)
+	struct expand_four_T *restrict e)
     __attribute__((nonnull));
-static wchar_t *expand_param_simple(const paramexp_T *p)
-    __attribute__((nonnull,malloc,warn_unused_result));
 static enum indextype_T parse_indextype(const wchar_t *indexstr)
     __attribute__((nonnull,pure));
 static wchar_t *trim_wstring(wchar_t *s, ssize_t startindex, ssize_t endindex)
@@ -89,7 +98,9 @@ static void match_each(void **restrict slist, const wchar_t *restrict pattern,
 static void subst_each(void **restrict slist, const wchar_t *pattern,
 	const wchar_t *subst, paramexptype_T type)
     __attribute__((nonnull));
-static void **concatenate_values(void **values)
+static wchar_t *concatenate_values(void **values, bool escape)
+    __attribute__((nonnull,malloc,warn_unused_result));
+static void **concatenate_values_into_array(void **values, bool escape)
     __attribute__((nonnull,malloc,warn_unused_result));
 static void subst_length_each(void **slist)
     __attribute__((nonnull));
@@ -107,11 +118,22 @@ static bool try_expand_brace_sequence(
 static bool has_leading_zero(const wchar_t *restrict s, bool *restrict sign)
     __attribute__((nonnull));
 
+static void fieldsplit_all(void **restrict valuelist, void **restrict splitlist,
+	plist_T *restrict dest)
+    __attribute__((nonnull));
 static void fieldsplit(wchar_t *restrict s, char *restrict split,
 	const wchar_t *restrict ifs, plist_T *restrict dest)
     __attribute__((nonnull));
-static void fieldsplit_all(void **restrict valuelist, void **restrict splitlist,
-	plist_T *restrict dest)
+static size_t skip_ifs(const wchar_t *s, const char *split,
+	bool escaped, const wchar_t *ifs)
+    __attribute__((nonnull,pure));
+static size_t skip_ifs_whitespaces(const wchar_t *s, const char *split,
+	bool escaped, const wchar_t *ifs)
+    __attribute__((nonnull,pure));
+static size_t skip_field(const wchar_t *s, const char *split,
+	bool escaped, const wchar_t *ifs)
+    __attribute__((nonnull,pure));
+static void add_empty_field(plist_T *dest, const wchar_t *p)
     __attribute__((nonnull));
 
 static inline void add_sq(
@@ -119,6 +141,10 @@ static inline void add_sq(
     __attribute__((nonnull));
 static wchar_t *escaped_wcspbrk(const wchar_t *s, const wchar_t *accept)
     __attribute__((nonnull));
+static wchar_t *escaped_remove(const wchar_t *s, const wchar_t *reject)
+    __attribute__((nonnull,malloc,warn_unused_result));
+static inline wchar_t *escaped_remove_free(wchar_t *s, const wchar_t *reject)
+    __attribute__((nonnull,malloc,warn_unused_result));
 
 static void glob_all(void **restrict patterns, plist_T *restrict list)
     __attribute__((nonnull));
@@ -199,12 +225,24 @@ bool expand_and_split_words(
     pl_init(&valuelist1);
     pl_init(&splitlist1);
 
+    struct expand_four_T expand;
+    expand.valuelist = &valuelist1;
+    expand.splitlist = &splitlist1;
+    wb_init(&expand.valuebuf);
+    sb_init(&expand.splitbuf);
+    expand.zeroword = false;
+
     /* four expansions (w -> list1) */
-    if (!expand_word(w, TT_SINGLE, false, &valuelist1, &splitlist1)) {
+    if (!expand_four(w, TT_SINGLE, true, false, false, &expand)) {
 	plfree(pl_toary(&valuelist1), free);
 	plfree(pl_toary(&splitlist1), free);
+	wb_destroy(&expand.valuebuf);
+	sb_destroy(&expand.splitbuf);
 	return false;
     }
+    assert(expand.valuebuf.length == expand.splitbuf.length);
+    pl_add(expand.valuelist, wb_towcs(&expand.valuebuf));
+    pl_add(expand.splitlist, sb_tostr(&expand.splitbuf));
 
     /* brace expansion (list1 -> list2) */
     if (shopt_braceexpand) {
@@ -220,39 +258,62 @@ bool expand_and_split_words(
     }
 
     /* field splitting (list2 -> list) */
+    size_t oldlength = list->length;
     fieldsplit_all(pl_toary(&valuelist2), pl_toary(&splitlist2), list);
+    assert(oldlength <= list->length);
+
+    /* empty field removal */
+    if (list->length - oldlength == 1) {
+	wchar_t *field = list->contents[oldlength];
+	if (field[0] == L'\0' ||
+		(expand.zeroword && wcscmp(field, L"\"\"") == 0)) {
+	    free(field);
+	    pl_remove(list, oldlength, 1);
+	}
+    }
+
+    /* quote removal */
+    for (size_t i = oldlength; i < list->length; i++)
+	list->contents[i] = escaped_remove_free(list->contents[i], L"\"\'");
+
     return true;
 }
 
 /* Expands a single word: the four expansions and quote removal.
  * This function doesn't perform brace expansion, field splitting, globbing and
  * unescaping.
+ * If `processquotes' is true, single- and double-quotations are recognized as
+ * quotes. Otherwise, they are treated like backslashed characters.
+ * If `escapeall' is true, the expanded words are all backslashed as if the
+ * entire expansion is quoted.
+ * If `processquotes' and `escapeall' are false, only backslashes not preceding
+ * any of $, `, \ are self-backslashed.
  * If successful, the resulting word is returned as a newly malloced string
  * that may include backslash escapes.
  * On error, an error message is printed and NULL is returned.
  * On error in a non-interactive shell, the shell exits. */
-wchar_t *expand_single(const wordunit_T *arg, tildetype_T tilde)
+wchar_t *expand_single(const wordunit_T *arg,
+	tildetype_T tilde, bool processquotes, bool escapeall)
 {
-    wchar_t *result;
     plist_T list;
     pl_init(&list);
 
-    if (!expand_word(arg, tilde, false, &list, NULL)) {
+    if (!expand_four_and_remove_quotes(
+		arg, tilde, processquotes, escapeall, &list)) {
 	maybe_exit_on_error();
 	plfree(pl_toary(&list), free);
 	return NULL;
     }
-    if (list.length != 1) {
-	/* concatenate multiple words to a single */
-	const wchar_t *ifs = getvar(L VAR_IFS);
-	wchar_t padding[] = { ifs != NULL ? ifs[0] : L' ', L'\0' };
-	result = joinwcsarray(list.contents, padding);
-	plfree(pl_toary(&list), free);
-    } else {
-	result = list.contents[0];
-	pl_destroy(&list);
-    }
-    return result;
+
+    return concatenate_values(pl_toary(&list), true);
+}
+
+/* Like `expand_single', but the result is unescaped (if successful). */
+wchar_t *expand_single_and_unescape(const wordunit_T *arg,
+	tildetype_T tilde, bool processquotes, bool escapeall)
+{
+    wchar_t *result = expand_single(arg, tilde, processquotes, escapeall);
+    return result == NULL ? NULL : unescapefree(result);
 }
 
 /* Expands a single word: the four expansions, glob, quote removal and unescape.
@@ -267,7 +328,7 @@ wchar_t *expand_single(const wordunit_T *arg, tildetype_T tilde)
  * On error in a non-interactive shell, the shell exits. */
 char *expand_single_with_glob(const wordunit_T *arg, tildetype_T tilde)
 {
-    wchar_t *exp = expand_single(arg, tilde);
+    wchar_t *exp = expand_single(arg, tilde, true, false);
     char *result;
 
     if (exp == NULL)
@@ -314,129 +375,79 @@ noglob:
     return result;
 }
 
-/* Performs expansions in a string: parameter/arithmetic/command expansions.
- * Brace expansion, field splitting and globbing are not performed.
- * If `esc' is true, backslashes preceding $, `, \ are removed. Otherwise,
- * no quotations are removed.
- * The result is a newly malloced string.
- * On error, an error message is printed and NULL is returned. */
-wchar_t *expand_string(const wordunit_T *w, bool esc)
-{
-    bool ok = true;
-    xwcsbuf_T buf;
-    wchar_t *s;
-
-    wb_init(&buf);
-    for (; w != NULL; w = w->next) {
-	switch (w->wu_type) {
-	case WT_STRING:
-	    for (const wchar_t *ss = w->wu_string; *ss != L'\0'; ss++) {
-		if (esc && ss[0] == L'\\' && ss[1] != L'\0'
-			&& wcschr(L"$`\\", ss[1]) != NULL) {
-		    ss++;
-		    if (*ss != L'\0')
-			wb_wccat(&buf, *ss);
-		} else {
-		    wb_wccat(&buf, *ss);
-		}
-	    }
-	    break;
-	case WT_PARAM:
-	    s = expand_param_simple(w->wu_param);
-	    goto cat_s;
-	case WT_CMDSUB:
-	    s = exec_command_substitution(&w->wu_cmdsub);
-	    goto cat_s;
-	case WT_ARITH:
-	    s = expand_single(w->wu_arith, TT_NONE);
-	    if (s != NULL)
-		s = evaluate_arithmetic(unescapefree(s));
-	cat_s:
-	    if (s != NULL) {
-		wb_catfree(&buf, s);
-	    } else {
-		ok = false;
-	    }
-	    break;
-	}
-    }
-    if (ok) {
-	return wb_towcs(&buf);
-    } else {
-	wb_destroy(&buf);
-	maybe_exit_on_error();
-	return NULL;
-    }
-}
-
 
 /********** Four Expansions **********/
 
 /* Performs the four expansions in the specified single word.
- * The four expansions are tilde expansion, parameter expansion, command
- * substitution and arithmetic expansion.
  * `w' is the word in which expansions occur.
  * `tilde' is type of tilde expansion that is performed.
- * If `quoted' is true, the expanded words are all backslashed as if the entire
- * expansion is quoted.
+ * If `processquotes' is true, single- and double-quotations are recognized as
+ * quotes. Otherwise, they are treated like backslashed characters.
+ * If `escapeall' is true, the expanded words are all backslashed as if the
+ * entire expansion is quoted.
+ * If `processquotes' and `escapeall' are false, only backslashes not preceding
+ * any of $, `, \ are self-backslashed.
  * The expanded word is added to `valuelist' as a newly malloced wide string.
- * The splittability string is added to `splitlist' if `splitlist' is non-NULL.
  * Single- or double-quoted characters are unquoted and backslashed.
- * In most cases, one string is added to each of `valuelist' and `splitlist'.
- * If the word contains "$@", the result may be any number of strings.
+ * In most cases, one string is added to `valuelist'. If the word contains "$@",
+ * however, any number of strings may be added.
  * The return value is true iff successful. */
-/* A splittability string is an array of Boolean values that specifies where
- * the word can be split in field splitting. The word can be split at the nth
- * character iff the nth value of the splittability string is non-zero. */
-bool expand_word(
-	const wordunit_T *restrict w, tildetype_T tilde, bool quoted,
-	plist_T *restrict valuelist, plist_T *restrict splitlist)
+bool expand_four_and_remove_quotes(
+	const wordunit_T *restrict w,
+	tildetype_T tilde, bool processquotes, bool escapeall,
+	plist_T *restrict valuelist)
 {
-    struct expand_word_T expand;
+    size_t oldlength = valuelist->length;
+    struct expand_four_T expand;
 
     expand.valuelist = valuelist;
     wb_init(&expand.valuebuf);
-    expand.splitlist = splitlist;
-    if (expand.splitlist != NULL)
-	sb_init(&expand.splitbuf);
-    expand.putempty = false;
+    expand.splitlist = NULL;
+    expand.zeroword = false;
 
-    bool ok = expand_word_inner(w, tilde, quoted, false, &expand);
+    bool ok = expand_four(w, tilde, processquotes, escapeall, false, &expand);
 
-    if (expand.splitlist != NULL)
-	assert(expand.valuebuf.length == expand.splitbuf.length);
-
-    /* A quoted empty word, if any, is added to the list here. It is indicated
-     * by the `putempty' flag that is set when a quote is found. */
-    if (expand.valuebuf.length > 0 || expand.putempty) {
-	pl_add(expand.valuelist, wb_towcs(&expand.valuebuf));
-	if (expand.splitlist != NULL)
-	    pl_add(expand.splitlist, sb_tostr(&expand.splitbuf));
-    } else {
+    /* remove empty word for "$@" if $# == 0 */
+    if (valuelist->length == oldlength && expand.zeroword &&
+	    wcscmp(expand.valuebuf.contents, L"\"\"") == 0)
 	wb_destroy(&expand.valuebuf);
-	if (expand.splitlist != NULL)
-	    sb_destroy(&expand.splitbuf);
-    }
+    else
+	pl_add(valuelist, wb_towcs(&expand.valuebuf));
+
+    /* quote removal */
+    for (size_t i = oldlength; i < valuelist->length; i++)
+	valuelist->contents[i] =
+	    escaped_remove_free(valuelist->contents[i], L"\"\'");
 
     return ok;
 }
 
 /* Performs the four expansions in the specified single word.
+ * The four expansions are tilde expansion, parameter expansion, command
+ * substitution, and arithmetic expansion.
  * `w' is the word in which expansions occur.
  * `tilde' specifies the type of tilde expansion that is performed.
- * If `quoted' is true, the expanded words are all backslashed as if the entire
- * expansion is quoted.
+ * If `processquotes' is true, single- and double-quotations are recognized as
+ * quotes. Otherwise, they are treated like backslashed characters.
+ * If `escapeall' is true, the expanded words are all backslashed as if the
+ * entire expansion is quoted.
+ * If `processquotes' and `escapeall' are false, only backslashes not preceding
+ * any of $, `, \ are self-backslashed.
  * `rec' must be true iff this expansion is part of another expansion.
  * `e->valuebuf' must be initialized before calling this function and is used to
  * expand the current word. If `w' expands to multiple words, the last word is
  * put in `e->valuebuf' and the others are inserted to `e->valuelist'.
  * The splittability strings are put in `e->splitbuf' and `e->splitlist'
  * accordingly if `e->splitlist' is non-NULL.
- * Single- or double-quoted characters are unquoted and backslashed.
+ * Single- and double-quotations remain in the resulting word. In addition,
+ * characters inside those quotations are backslashed.
  * The return value is true iff successful. */
-bool expand_word_inner(const wordunit_T *restrict w,
-	tildetype_T tilde, bool quoted, bool rec,
-	struct expand_word_T *restrict e)
+/* A splittability string is an array of Boolean values that specifies where
+ * the word can be split in field splitting. The word can be split at the nth
+ * character iff the nth value of the splittability string is non-zero. */
+bool expand_four(const wordunit_T *restrict w,
+	tildetype_T tilde, bool processquotes, bool escapeall, bool rec,
+	struct expand_four_T *restrict e)
 {
     bool ok = true;
     bool indq = false;  /* in a double quote? */
@@ -444,12 +455,7 @@ bool expand_word_inner(const wordunit_T *restrict w,
     const wchar_t *ss;
     wchar_t *s;
 
-#define FILL_SBUF(c)                                            \
-    do {                                                        \
-	if (e->splitlist != NULL)                               \
-	    sb_ccat_repeat(&e->splitbuf, c,                     \
-		    e->valuebuf.length - e->splitbuf.length);   \
-    } while (false)
+#define FILL_SBUF(s) fill_splitbuf(e, !indq && !escapeall && (s));
 #define FILL_SBUF_SPLITTABLE   FILL_SBUF(true)
 #define FILL_SBUF_UNSPLITTABLE FILL_SBUF(false)
 
@@ -467,19 +473,32 @@ bool expand_word_inner(const wordunit_T *restrict w,
 	    while (*ss != L'\0') {
 		switch (*ss) {
 		case L'"':
+		    if (!processquotes)
+			goto escape;
 		    indq = !indq;
-		    e->putempty |= indq;
+		    wb_wccat(&e->valuebuf, L'"');
+		    FILL_SBUF_UNSPLITTABLE;
 		    break;
 		case L'\'':
-		    if (indq)
-			goto default_case;
-		    e->putempty = true;
+		    if (!processquotes || indq)
+			goto escape;
+		    wb_wccat(&e->valuebuf, L'\'');
 		    add_sq(&ss, &e->valuebuf, true);
+		    wb_wccat(&e->valuebuf, L'\'');
 		    FILL_SBUF_UNSPLITTABLE;
 		    break;
 		case L'\\':
+		    if (!processquotes) {
+			if (!escapeall) {
+			    wchar_t c = ss[1];
+			    if (c == L'$' || c == L'`' || c == L'\\')
+				ss++;
+			}
+			goto escape;
+		    }
+
 		    if (indq && wcschr(CHARS_ESCAPABLE, ss[1]) == NULL) {
-			goto default_case;
+			goto escape;
 		    } else {
 			wb_wccat(&e->valuebuf, L'\\');
 			if (*++ss != L'\0')
@@ -500,32 +519,33 @@ bool expand_word_inner(const wordunit_T *restrict w,
 			continue;
 		    }
 		    /* falls thru! */
-		default:  default_case:
-		    if (indq || quoted)
+		default:
+		    if (indq || escapeall)
+escape:
 			wb_wccat(&e->valuebuf, L'\\');
 		    wb_wccat(&e->valuebuf, *ss);
-		    FILL_SBUF(rec && !indq && !quoted);
+		    FILL_SBUF(rec);
 		    break;
 		}
 		ss++;
 	    }
 	    break;
 	case WT_PARAM:
-	    if (!expand_param(w->wu_param, indq || quoted, e))
+	    if (!expand_param(w->wu_param, indq || escapeall, e))
 		ok = false;
 	    break;
 	case WT_CMDSUB:
 	    s = exec_command_substitution(&w->wu_cmdsub);
 	    goto cat_s;
 	case WT_ARITH:
-	    s = expand_single(w->wu_arith, TT_NONE);
+	    s = expand_single_and_unescape(w->wu_arith, TT_NONE, true, false);
 	    if (s != NULL)
-		s = evaluate_arithmetic(unescapefree(s));
+		s = evaluate_arithmetic(s);
 cat_s:
 	    if (s != NULL) {
 		wb_catfree(&e->valuebuf, escapefree(s,
-			    (indq || quoted) ? NULL : CHARS_ESCAPED));
-		FILL_SBUF(!(indq || quoted));
+			    (indq || escapeall) ? NULL : CHARS_ESCAPED));
+		FILL_SBUF_SPLITTABLE;
 	    } else {
 		ok = false;
 	    }
@@ -533,9 +553,23 @@ cat_s:
 	}
     }
 
+#undef FILL_SBUF_UNSPLITTABLE
+#undef FILL_SBUF_SPLITTABLE
+#undef FILL_SBUF
+
     return ok;
 }
 
+/* Appends to `e->splitbuf' as many `splittable' as needed to match the length
+ * with `e->valuebuf'. */
+void fill_splitbuf(struct expand_four_T *e, bool splittable)
+{
+    if (e->splitlist == NULL)
+	return;
+    sb_ccat_repeat(
+	    &e->splitbuf, splittable, e->valuebuf.length - e->splitbuf.length);
+}
+
 /* Performs tilde expansion.
  * `ss' is a pointer to a pointer to the tilde character. The pointer is
  * increased so that it points to the character right after the expanded string.
@@ -611,7 +645,7 @@ finish:
  * The result is put in `e'.
  * Returns true iff successful. */
 bool expand_param(const paramexp_T *restrict p, bool indq,
-	struct expand_word_T *restrict e)
+	struct expand_four_T *restrict e)
 {
     /* parse indices first */
     ssize_t startindex, endindex;
@@ -619,7 +653,8 @@ bool expand_param(const paramexp_T *restrict p, bool indq,
     if (p->pe_start == NULL) {
 	startindex = 0, endindex = SSIZE_MAX, indextype = IDX_NONE;
     } else {
-	wchar_t *start = expand_single(p->pe_start, TT_NONE);
+	wchar_t *start = expand_single_and_unescape(
+		p->pe_start, TT_NONE, true, false);
 	if (start == NULL)
 	    return false;
 	indextype = parse_indextype(start);
@@ -636,7 +671,8 @@ bool expand_param(const paramexp_T *restrict p, bool indq,
 	    if (p->pe_end == NULL) {
 		endindex = (startindex == -1) ? SSIZE_MAX : startindex;
 	    } else {
-		wchar_t *end = expand_single(p->pe_end, TT_NONE);
+		wchar_t *end = expand_single_and_unescape(
+			p->pe_end, TT_NONE, true, false);
 		if (end == NULL || !evaluate_index(end, &endindex))
 		    return false;
 	    }
@@ -656,7 +692,8 @@ bool expand_param(const paramexp_T *restrict p, bool indq,
     if (p->pe_type & PT_NEST) {
 	plist_T plist;
 	pl_init(&plist);
-	if (!expand_word(p->pe_nest, TT_NONE, true, &plist, NULL)) {
+	if (!expand_four_and_remove_quotes(
+		    p->pe_nest, TT_NONE, true, true, &plist)) {
 	    plfree(pl_toary(&plist), free);
 	    return false;
 	}
@@ -773,7 +810,7 @@ treat_array:
 	if (unset) {
 subst:
 	    plfree(values, free);
-	    return expand_word_inner(p->pe_subst, TT_SINGLE, indq, true, e);
+	    return expand_four(p->pe_subst, TT_SINGLE, true, indq, true, e);
 	}
 	break;
     case PT_ASSIGN:
@@ -795,10 +832,10 @@ subst:
 			p->pe_name);
 		return false;
 	    }
-	    subst = expand_single(p->pe_subst, TT_SINGLE);
+	    subst = expand_single_and_unescape(
+		    p->pe_subst, TT_SINGLE, true, false);
 	    if (subst == NULL)
 		return false;
-	    subst = unescapefree(subst);
 	    if (v.type != GV_ARRAY) {
 		assert(v.type == GV_NOTFOUND || v.type == GV_SCALAR);
 		if (!set_variable(
@@ -838,7 +875,7 @@ subst:
     wchar_t *match;
     switch (p->pe_type & PT_MASK) {
     case PT_MATCH:
-	match = expand_single(p->pe_match, TT_SINGLE);
+	match = expand_single(p->pe_match, TT_SINGLE, true, false);
 	if (match == NULL) {
 	    plfree(values, free);
 	    return false;
@@ -847,15 +884,14 @@ subst:
 	free(match);
 	break;
     case PT_SUBST:
-	match = expand_single(p->pe_match, TT_SINGLE);
-	subst = expand_single(p->pe_subst, TT_SINGLE);
+	match = expand_single(p->pe_match, TT_SINGLE, true, false);
+	subst = expand_single_and_unescape(p->pe_subst, TT_SINGLE, true, false);
 	if (match == NULL || subst == NULL) {
 	    free(match);
 	    free(subst);
 	    plfree(values, free);
 	    return false;
 	}
-	subst = unescapefree(subst);
 	subst_each(values, match, subst, p->pe_type);
 	free(match);
 	free(subst);
@@ -864,7 +900,7 @@ subst:
 
     /* concatenate the elements of `values' */
     if (concat)
-	values = concatenate_values(values);
+	values = concatenate_values_into_array(values, false);
 
     /* PT_NUMBER */
     if (p->pe_type & PT_NUMBER)
@@ -872,7 +908,7 @@ subst:
 
     /* concatenate the elements of `values' */
     if (!indq)
-	values = concatenate_values(values);
+	values = concatenate_values_into_array(values, false);
 
     /* backslash escape */
     for (size_t i = 0; values[i] != NULL; i++)
@@ -880,11 +916,11 @@ subst:
 
     /* add the elements of `values' to `e->valuelist' */
     if (values[0] == NULL) {
-	e->putempty = false;
+	e->zeroword = true;
     } else {
 	/* add the first element */
 	wb_catfree(&e->valuebuf, values[0]);
-	FILL_SBUF(!indq);
+	fill_splitbuf(e, !indq);
 	if (values[1] != NULL) {
 	    pl_add(e->valuelist, wb_towcs(&e->valuebuf));
 	    if (e->splitlist != NULL)
@@ -904,7 +940,7 @@ subst:
 	    wb_initwith(&e->valuebuf, values[i]);
 	    if (e->splitlist != NULL) {
 		sb_init(&e->splitbuf);
-		FILL_SBUF(!indq);
+		fill_splitbuf(e, !indq);
 	    }
 	}
     }
@@ -913,40 +949,6 @@ subst:
     return true;
 }
 
-/* Performs parameter expansion and returns the resulting word.
- * If multiple words result, they are concatenated into a single string, each
- * separated by a space.
- * If successful, returns a newly-malloced wide string.
- * On error, returns NULL. */
-wchar_t *expand_param_simple(const paramexp_T *p)
-{
-    plist_T valuelist;
-    struct expand_word_T expand;
-
-    expand.valuelist = pl_init(&valuelist);
-    wb_init(&expand.valuebuf);
-    expand.splitlist = NULL;
-    expand.putempty = false;
-
-    bool ok = expand_param(p, false, &expand);
-
-    /* A quoted empty word, if any, is added to the list here. It is indicated
-     * by the `putempty' flag, which is set when a quote is found. */
-    if (expand.valuebuf.length > 0 || expand.putempty) {
-	pl_add(expand.valuelist, wb_towcs(&expand.valuebuf));
-    } else {
-	wb_destroy(&expand.valuebuf);
-    }
-
-    for (size_t i = 0; i < valuelist.length; i++)
-	valuelist.contents[i] = unescapefree(valuelist.contents[i]);
-
-    void **results = pl_toary(expand.valuelist);
-    wchar_t *result = ok ? joinwcsarray(results, L" ") : NULL;
-    plfree(results, free);
-    return result;
-}
-
 /* Returns IDX_ALL, IDX_CONCAT, IDX_NUMBER if `indexstr' is L"@", L"*",
  * L"#" respectively. Otherwise returns IDX_NONE. */
 enum indextype_T parse_indextype(const wchar_t *indexstr)
@@ -1033,9 +1035,9 @@ void **trim_array(void **a, ssize_t startindex, ssize_t endindex)
 void print_subst_as_error(const paramexp_T *p)
 {
     if (p->pe_subst != NULL) {
-	wchar_t *subst = expand_single(p->pe_subst, TT_SINGLE);
+	wchar_t *subst = expand_single_and_unescape(
+		p->pe_subst, TT_SINGLE, true, false);
 	if (subst != NULL) {
-	    subst = unescapefree(subst);
 	    if (p->pe_type & PT_NEST)
 		xerror(0, "%ls", subst);
 	    else
@@ -1120,25 +1122,37 @@ void subst_each(void **restrict slist, const wchar_t *pattern,
 /* Concatenates the wide strings in the specified array.
  * Array `*values' must be a NULL-terminated array of pointers to wide strings.
  * The strings are concatenated into one, each separated by the first $IFS
- * character.
+ * character. The separators are backslashed if `escape' is true.
  * The array and its element strings are all freed in this function.
- * The return value is a pointer to a newly malloced NULL-terminated array that
- * contains exactly one element: a pointer to the newly malloced wide string
- * that is the result of concatenation. */
-void **concatenate_values(void **values)
+ * The return value is a pointer to the newly malloced wide string that is the
+ * result of concatenation. */
+wchar_t *concatenate_values(void **values, bool escape)
+{
+    wchar_t *first = values[0];
+    if (first != NULL && values[1] == NULL) {
+	// no actual concatenation needed
+	free(values);
+	return first;
+    }
+
+    const wchar_t *ifs = getvar(L VAR_IFS);
+    wchar_t padding[] = { L'\\', ifs != NULL ? ifs[0] : L' ', L'\0' };
+    wchar_t *result = joinwcsarray(values, escape ? padding : &padding[1]);
+    plfree(values, free);
+    return result;
+}
+
+/* Like `concatenate_values', but returns a pointer to a newly malloced
+ * NULL-terminated array containing the concatenated string. */
+void **concatenate_values_into_array(void **values, bool escape)
 {
-    /* If the array contains just one string, simply return it. */
     if (values[0] != NULL && values[1] == NULL)
 	return values;
 
-    const wchar_t *ifs = getvar(L VAR_IFS);
-    wchar_t padding[] = { ifs != NULL ? ifs[0] : L' ', L'\0' };
-    wchar_t *result = joinwcsarray(values, padding);
-    plfree(values, free);
-    values = xmallocn(2, sizeof *values);
-    values[0] = result;
-    values[1] = NULL;
-    return values;
+    void **results = xmallocn(2, sizeof *values);
+    results[0] = concatenate_values(values, escape);
+    results[1] = NULL;
+    return results;
 }
 
 /* Substitutes each string in the specified array with a string that contains
@@ -1287,7 +1301,7 @@ bool try_expand_brace_sequence(
 	return false;
     startlen = has_leading_zero(c, &sign) ? (dotp - c) : 0;
     errno = 0;
-    start = wcstol(c, &c, 0);
+    start = wcstol(c, &c, 10);
     if (errno != 0 || c != dotp)
 	return false;
 
@@ -1300,7 +1314,7 @@ bool try_expand_brace_sequence(
 	return false;
     endlen = has_leading_zero(c, &sign) ? (dotbracep - c) : 0;
     errno = 0;
-    end = wcstol(c, &c, 0);
+    end = wcstol(c, &c, 10);
     if (errno != 0 || c != dotbracep)
 	return false;
 
@@ -1312,7 +1326,7 @@ bool try_expand_brace_sequence(
 	if (bracep == NULL || c == bracep)
 	    return false;
 	errno = 0;
-	delta = wcstol(c, &c, 0);
+	delta = wcstol(c, &c, 10);
 	if (delta == 0 || errno != 0 || c != bracep)
 	    return false;
     } else {
@@ -1386,76 +1400,6 @@ bool has_leading_zero(const wchar_t *restrict s, bool *restrict sign)
 
 /********** Field Splitting **********/
 
-/* Performs field splitting.
- * `s' is the word to split and freed in this function.
- * `split' is the splittability string corresponding to `s' and also freed.
- * The results are added to `dest' as newly-malloced wide strings.
- * `ifs' must not be NULL.
- * The word is split at characters that are contained in `ifs' and whose
- * corresponding character in the splittability string is non-zero. */
-/* Field are split at characters that are contained in `ifs'. If an empty field
- * is split by white spaces, the field is ignored.
- * Split examples (assuming `ifs' = L" -")
- *   " abc 123 "         ->   "abc" "123"
- *   "  abc  123  "      ->   "abc" "123"
- *   "-abc-123-"         ->   "" "abc" "123" ""
- *   " - abc - 123 - "   ->   "" "abc" "123" ""
- *   "abc--123"          ->   "abc" "" "123"
- *   "abc - - 123"       ->   "abc" "" "123"
- */
-void fieldsplit(wchar_t *restrict s, char *restrict split,
-	const wchar_t *restrict ifs, plist_T *restrict dest)
-{
-    size_t index = 0;
-    size_t lwstart = 0, lwend;  /* start/end index of last word */
-
-    for (;;) {
-	lwend = index;
-	if (s[index] == L'\\')
-	    index++;
-	if (s[index] == L'\0')
-	    break;
-	if (split[index] && wcschr(ifs, s[index]) != NULL) {
-	    /* the character is in `ifs', so do splitting */
-	    bool splitatnonspace = false, nonspace = false;
-	    if (lwstart < lwend)
-		pl_add(dest, xwcsndup(s + lwstart, lwend - lwstart));
-	    else
-		splitatnonspace = true;
-	    do {
-		if (!iswspace(s[index])) {
-		    if (splitatnonspace)
-			pl_add(dest, xwcsdup(L""));
-		    splitatnonspace = true;
-		    nonspace = true;
-		}
-		index++;
-		if (s[index] == L'\\')
-		    index++;
-		if (s[index] == L'\0') {
-		    if (nonspace)
-			pl_add(dest, xwcsdup(L""));
-		    break;
-		}
-	    } while (split[index] && wcschr(ifs, s[index]) != NULL);
-	    lwstart = index;
-	} else {
-	    index++;
-	}
-    }
-    if (lwstart < lwend || lwend == 0) {
-	/* if we have some leftover or the string is empty at all, add it. */
-	if (lwstart > 0) {
-	    xwcsbuf_T buf;
-	    s = wb_towcs(wb_remove(wb_initwith(&buf, s), 0, lwstart));
-	}
-	pl_add(dest, s);
-    } else {
-	free(s);
-    }
-    free(split);
-}
-
 /* Performs field splitting.
  * `valuelist' is a NULL-terminated array of pointers to wide strings to split.
  * `splitlist' is an array of pointers to corresponding splittability strings.
@@ -1478,73 +1422,165 @@ void fieldsplit_all(void **restrict valuelist, void **restrict splitlist,
     free(splitlist);
 }
 
-/* Extracts a field starting at `**sp'.
- * Pointer `*sp' is advanced to the first character of the next field or
- * to the end of string (L'\0') if none. If `**sp' is already L'\0', an empty
- * string is returned.
- * If `ifs' is NULL, the default separator is used.
- * If `noescape' is false, backslashes are treaded as escape characters.
- * The returned string is a newly malloced string. */
-wchar_t *split_next_field(const wchar_t **sp, const wchar_t *ifs, bool noescape)
+/* Performs field splitting.
+ * `s' is the word to split and freed in this function.
+ * `split' is the splittability string corresponding to `s' and also freed.
+ * The results are added to `dest' as newly-malloced wide strings.
+ * `ifs' must not be NULL. */
+void fieldsplit(wchar_t *restrict s, char *restrict split,
+	const wchar_t *restrict ifs, plist_T *restrict dest)
 {
-    size_t i = 0, length = 0;
-    const wchar_t *s = *sp;
-    wchar_t *result;
+    plist_T fields;
 
-    if (ifs == NULL)
-	ifs = DEFAULT_IFS;
-    while (*s != L'\0' && wcschr(ifs, *s) != NULL && iswspace(*s))
-	s++;
-    if (*s != L'\0' && wcschr(ifs, *s) != NULL && !iswspace(*s)) {
-	assert(length == 0);
-	i = 1;
-	goto end;
-    }
-    while (s[i] != L'\0') {
-	if (!noescape && s[i] == L'\\') {
-	    i++;
-	    if (s[i] == L'\0') {
-		length = i;
-		break;
-	    }
-	    i++;
-	} else if (wcschr(ifs, s[i]) != NULL) {
-	    length = i;
-	    do {
-		if (!iswspace(s[i])) {
-		    i++;
-		    while (wcschr(ifs, s[i]) != NULL && iswspace(s[i]))
-			i++;
-		    break;
-		}
-	    } while (wcschr(ifs, s[++i]) != NULL);
-	    break;
-	} else {
-	    i++;
-	    length = i;
-	}
+    pl_init(&fields);
+    extract_fields(s, split, true, ifs, &fields);
+    assert(fields.length % 2 == 0);
+
+    for (size_t i = 0; i < fields.length; i += 2) {
+	const wchar_t *start = fields.contents[i], *end = fields.contents[i+1];
+	pl_add(dest, xwcsndup(start, end - start));
     }
 
-end:
-    result = xwcsndup(s, length);
-    if (!noescape)
-	result = unescapefree(result);
-    *sp = s + i;
-    return result;
+    pl_destroy(&fields);
+    free(s);
+    free(split);
 }
 
-/* Removes IFS white spaces at the end of the string `s' by replacing them with
- * null characters. The default IFS is used if `ifs' is NULL. */
-void trim_trailing_spaces(wchar_t *restrict s, const wchar_t *restrict ifs)
+/* Extracts fields from a string.
+ * `s' is the word to split.
+ * `split' is the splittability string corresponding to `s'. It must be at least
+ * as long as `wcslen(s)'.
+ * If `escaped' is true, backslashes in `s' are treated as escapes. But
+ * backslashes do not prevent splitting.
+ * `ifs' must not be NULL.
+ *
+ * The results are appended to `dest'. If n fields are found, 2n pointers are
+ * appended to `dest'. The first pointer points to the first character of the
+ * first field in `s'. The second to the character past the last character of
+ * the first field. The third to the first character of the second field. And so
+ * on.
+ *
+ * The word is split at characters that are contained in `ifs' and whose
+ * corresponding character in the splittability string is non-zero. Refer to
+ * POSIX for how whitespaces are treated in field splitting.
+ *
+ * If an IFS non-whitespace delimits an empty field, the field is assumed just
+ * before the non-whitespace delimiter. The empty last field is removed if
+ * `shopt_emptylastfield' is false.
+ *
+ * The return value is a pointer to the end of the input string (but before
+ * trailing IFS whitespaces). */
+/* Split examples (assuming `ifs' = L" -" and `shopt_emptylastfield' is true)
+ *   ""                  ->   ""
+ *   "  "                ->   ""
+ *   " abc 123 "         ->   "abc" "123"
+ *   "  abc  123  "      ->   "abc" "123"
+ *   "-abc-123-"         ->   "" "abc" "123" ""
+ *   " - abc - 123 - "   ->   "" "abc" "123" ""
+ *   "abc--123"          ->   "abc" "" "123"
+ *   "abc - - 123"       ->   "abc" "" "123"
+ */
+wchar_t *extract_fields(const wchar_t *restrict s, const char *restrict split,
+	bool escaped, const wchar_t *restrict ifs, plist_T *restrict dest)
 {
-    if (ifs == NULL)
-	ifs = DEFAULT_IFS;
-    for (size_t i = wcslen(s); i-- > 0; ) {
-	if (wcschr(ifs, s[i]) != NULL && iswspace(s[i]))
-	    s[i] = L'\0';
-	else
-	    break;
+    size_t index = 0;
+    size_t ifswhitestartindex;
+    size_t oldlen = dest->length;
+
+    /* true when the currently skipping IFS whitespaces immediately follow a
+     * previously split field. */
+    bool afterfield = false;
+
+    for (;;) {
+	ifswhitestartindex = index;
+	index += skip_ifs_whitespaces(&s[index], &split[index], escaped, ifs);
+
+	/* extract next field, if any */
+	size_t fieldstartindex = index;
+	index += skip_field(&s[index], &split[index], escaped, ifs);
+	if (index != fieldstartindex) {
+	    pl_add(pl_add(dest, &s[fieldstartindex]), &s[index]);
+	    afterfield = true;
+	    continue;
+	}
+
+	/* Now the current char is either null or a IFS non-whitespace. */
+
+	if (!afterfield)
+	    add_empty_field(dest, &s[index]);
+
+	/* skip (only one) IFS non-whitespace */
+	size_t ifsstartindex = index;
+	index += skip_ifs(&s[index], &split[index], escaped, ifs);
+	if (index != ifsstartindex) {
+	    afterfield = false;
+	    continue;
+	}
+
+	/* Now the current char is null. We're done. */
+	break;
     }
+
+    /* remove the empty last field */
+    size_t newlen = dest->length;
+    if (!shopt_emptylastfield && newlen - oldlen >= 2 * 2 &&
+	    dest->contents[newlen - 2] == dest->contents[newlen - 1])
+	pl_remove(dest, newlen - 2, 2);
+
+    assert(dest->length - oldlen >= 2);
+    return (wchar_t *) &s[ifswhitestartindex];
+}
+
+/* If `*s' is a (possibly escaped if `escaped') IFS character, returns the
+ * number of characters to skip it. Otherwise returns zero. */
+size_t skip_ifs(const wchar_t *s, const char *split,
+	bool escaped, const wchar_t *ifs)
+{
+    size_t i = 0;
+    if (escaped && s[i] == L'\\')
+	i++;
+    if (s[i] == L'\0')
+	return 0;
+    if (split[i] && wcschr(ifs, s[i]) != NULL)
+	return i + 1;
+    else
+	return 0;
+}
+
+/* Returns the length of IFS whitespace sequence starting at `*s'. */
+size_t skip_ifs_whitespaces(const wchar_t *s, const char *split,
+	bool escaped, const wchar_t *ifs)
+{
+    size_t total = 0;
+    for (;;) {
+	size_t current = skip_ifs(&s[total], &split[total], escaped, ifs);
+	if (current == 0 || !iswspace(s[total + current - 1]))
+	    return total;
+	total += current;
+    }
+}
+
+/* Returns the length of a field starting at `*s'. */
+size_t skip_field(const wchar_t *s, const char *split,
+	bool escaped, const wchar_t *ifs)
+{
+    size_t index = 0;
+    for (;;) {
+	size_t saveindex = index;
+	if (escaped && s[index] == L'\\')
+	    index++;
+	if (s[index] == L'\0')
+	    return saveindex;
+	if (split[index] && wcschr(ifs, s[index]) != NULL)
+	    return saveindex;
+	index++;
+    }
+}
+
+void add_empty_field(plist_T *dest, const wchar_t *p)
+{
+    pl_add(dest, p);
+    pl_add(dest, p);
 }
 
 
@@ -1699,6 +1735,33 @@ wchar_t *escaped_wcspbrk(const wchar_t *s, const wchar_t *accept)
     return NULL;
 }
 
+/* Removes characters in `reject' from `s'.
+ * Backslash escapes in `s' are recognized. Escapes and escaped characters are
+ * kept in the result.
+ * The result is a newly malloced string. */
+wchar_t *escaped_remove(const wchar_t *s, const wchar_t *reject)
+{
+    xwcsbuf_T result;
+    wb_init(&result);
+    for (;;) {
+	const wchar_t *rejectchar = escaped_wcspbrk(s, reject);
+	if (rejectchar == NULL)
+	    break;
+	wb_ncat_force(&result, s, rejectchar - s);
+	s = rejectchar + 1;
+    }
+    wb_cat(&result, s);
+    return wb_towcs(&result);
+}
+
+/* Like `escaped_remove', but frees `s' before returning the result. */
+wchar_t *escaped_remove_free(wchar_t *s, const wchar_t *reject)
+{
+    wchar_t *result = escaped_remove(s, reject);
+    free(s);
+    return result;
+}
+
 
 /********** File Name Expansion (Glob) **********/
 
@@ -1766,9 +1829,7 @@ wchar_t *parse_and_expand_string(const wchar_t *s, const char *name, bool esc)
     parseparam_T info = {
 	.print_errmsg = true,
 	.enable_verbose = false,
-#if YASH_ENABLE_ALIAS
 	.enable_alias = true,
-#endif
 	.filename = name,
 	.lineno = 1,
 	.input = input_wcs,
@@ -1780,7 +1841,7 @@ wchar_t *parse_and_expand_string(const wchar_t *s, const char *name, bool esc)
 
     if (!parse_string(&info, &word))
 	return NULL;
-    result = expand_string(word, esc);
+    result = expand_single_and_unescape(word, TT_NONE, false, !esc);
     wordfree(word);
     return result;
 }
