@@ -1,6 +1,6 @@
 /* Yash: yet another shell */
 /* input.c: functions for input of command line */
-/* (C) 2007-2015 magicant */
+/* (C) 2007-2016 magicant */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -95,19 +95,26 @@ inputresult_T read_input(
 	xwcsbuf_T *buf, struct input_file_info_T *info, bool trap)
 {
     size_t initlen = buf->length;
-    bool ok = true;
-
-    if (trap) {
-	handle_signals();
-	reset_sigint();
-    }
+    inputresult_T status = INPUT_EOF;
 
     for (;;) {
 	if (info->bufpos >= info->bufmax) {
 read_input:  /* if there's nothing in the buffer, read the next input */
-	    ok = wait_for_input(info->fd, trap, -1);
-	    if (!ok)
-		goto end;
+	    switch (wait_for_input(info->fd, trap, -1)) {
+		case W_READY:
+		    break;
+		case W_TIMED_OUT:
+		    assert(false);
+		case W_INTERRUPTED:
+		    // Ignore interruption and continue reading, because:
+		    //  1) POSIX does not require to do so, and
+		    //  2) the buffer for canonical-mode editing cannot be
+		    //     controlled from the shell.
+		    goto read_input;
+		case W_ERROR:
+		    status = INPUT_ERROR;
+		    goto end;
+	    }
 
 	    ssize_t readcount = read(info->fd, info->buf, info->bufsize);
 	    if (readcount < 0) switch (errno) {
@@ -151,14 +158,12 @@ read_input:  /* if there's nothing in the buffer, read the next input */
 
 error:
     xerror(errno, Ngt("cannot read input"));
-    ok = false;
+    status = INPUT_ERROR;
 end:
     if (initlen != buf->length)
 	return INPUT_OK;
-    else if (ok)
-	return INPUT_EOF;
     else
-	return INPUT_ERROR;
+	return status;
 }
 
 /* An input function that prints a prompt and reads input.
