@@ -1,6 +1,6 @@
 /* Yash: yet another shell */
 /* sig.c: signal handling */
-/* (C) 2007-2012 magicant */
+/* (C) 2007-2016 magicant */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -56,7 +56,7 @@
 /* About the shell's signal handling:
  *
  * Yash always catches SIGCHLD.
- * When job control is active, SIGTSTP is ignored.
+ * When job control is active, SIGTTIN, SIGTTOU, and SIGTSTP are ignored.
  * If the shell is interactive, SIGTERM and SIGQUIT are ignored and SIGINT and
  * SIGWINCH are caught.
  * Trapped signals are also caught.
@@ -89,6 +89,7 @@ static void sig_handler(int signum);
 static void handle_sigchld(void);
 static void set_trap(int signum, const wchar_t *command);
 static bool is_ignored(int signum);
+static void banish_phantoms(void);
 #if YASH_ENABLE_LINEEDIT
 # ifdef SIGWINCH
 static inline void handle_sigwinch(void);
@@ -254,6 +255,13 @@ static volatile sig_atomic_t rtsignal_received[RTSIZE];
 static wchar_t *rttrap_command[RTSIZE];
 #endif
 
+/* If true, the contents of `trap_command` and `rttrap_command` are all
+ * phantom; that is, non-empty trap commands in those arrays should be treated
+ * as "default". Phantoms are used to remember previous trap commands after
+ * they were actually reset to "default".
+ * If false, trap commands are real. */
+static bool is_phantom = false;
+
 /* The signal mask the shell inherited on invocation.
  * This mask is inherited by commands the shell invokes.
  * When a signal's trap is set, the signal is removed from this mask. */
@@ -285,8 +293,8 @@ static volatile sig_atomic_t sigwinch_received;
 
 /* true iff SIGCHLD is handled. */
 static bool main_handler_set = false;
-/* true iff SIGTSTP is ignored. */
-static bool job_handler_set = false;
+/* true iff SIGTTIN, SIGTTOU, and SIGTSTP are ignored. */
+static bool job_handlers_set = false;
 /* true iff SIGTERM, SIGINT, SIGQUIT and SIGWINCH are ignored/handled. */
 static bool interactive_handlers_set = false;
 
@@ -306,8 +314,10 @@ void set_signals(void)
 {
     sigset_t block = trapped_signals;
 
-    if (!job_handler_set && doing_job_control_now) {
-	job_handler_set = true;
+    if (!job_handlers_set && doing_job_control_now) {
+	job_handlers_set = true;
+	set_special_handler(SIGTTIN, SIG_IGN);
+	set_special_handler(SIGTTOU, SIG_IGN);
 	set_special_handler(SIGTSTP, SIG_IGN);
     }
 
@@ -339,8 +349,10 @@ void set_signals(void)
  * If `leave' is false, the setting for SIGCHLD are not restored. */
 void restore_signals(bool leave)
 {
-    if (job_handler_set) {
-	job_handler_set = false;
+    if (job_handlers_set) {
+	job_handlers_set = false;
+	reset_special_handler(SIGTTIN, SIG_IGN, leave);
+	reset_special_handler(SIGTTOU, SIG_IGN, leave);
 	reset_special_handler(SIGTSTP, SIG_IGN, leave);
     }
     if (interactive_handlers_set) {
@@ -364,15 +376,19 @@ void restore_signals(bool leave)
     }
 }
 
-/* Re-sets the signal handler for SIGTSTP according to the current
- * `doing_job_control_now' and `job_handler_set'. */
+/* Re-sets the signal handler for SIGTTIN, SIGTTOU, and SIGTSTP according to the
+ * current `doing_job_control_now' and `job_handlers_set'. */
 void reset_job_signals(void)
 {
-    if (doing_job_control_now && !job_handler_set) {
-	job_handler_set = true;
+    if (doing_job_control_now && !job_handlers_set) {
+	job_handlers_set = true;
+	set_special_handler(SIGTTIN, SIG_IGN);
+	set_special_handler(SIGTTOU, SIG_IGN);
 	set_special_handler(SIGTSTP, SIG_IGN);
-    } else if (!doing_job_control_now && job_handler_set) {
-	job_handler_set = false;
+    } else if (!doing_job_control_now && job_handlers_set) {
+	job_handlers_set = false;
+	reset_special_handler(SIGTTIN, SIG_IGN, false);
+	reset_special_handler(SIGTTOU, SIG_IGN, false);
 	reset_special_handler(SIGTSTP, SIG_IGN, false);
     }
 }
@@ -386,7 +402,7 @@ void reset_job_signals(void)
 void set_special_handler(int signum, void (*handler)(int signum))
 {
     const wchar_t *trap = trap_command[sigindex(signum)];
-    if (trap != NULL && trap[0] != L'\0')
+    if (!is_phantom && trap != NULL && trap[0] != L'\0')
 	return;  /* The signal handler has already been set. */
 
     struct sigaction action, oldaction;
@@ -562,15 +578,15 @@ int wait_for_sigchld(bool interruptible, bool return_on_trap)
  * `handle_sigchld' and `handle_sigwinch' are called to handle SIGCHLD and
  * SIGWINCH that are caught while waiting.
  * If `trap' is true, traps are also handled while waiting and the
- * `sigint_received' flag is cleared when this function returns.
+ * `sigint_received' flag is cleared when this function returns. The return
+ * value can be W_INTERRUPTED only if `trap' is true. Any SIGINT received before
+ * entering this function is ignored.
  * The maximum time length of wait is specified by `timeout' in milliseconds.
  * If `timeout' is negative, the wait time is unlimited.
  * If the wait is interrupted by a signal, this function will re-wait for the
  * specified timeout, which means that this function may wait for a time length
- * longer than the specified timeout.
- * This function returns true if the input is ready or false if an error
- * occurred or it timed out. */
-bool wait_for_input(int fd, bool trap, int timeout)
+ * longer than the specified timeout. */
+enum wait_for_input_T wait_for_input(int fd, bool trap, int timeout)
 {
     sigset_t ss;
     struct timespec to;
@@ -579,9 +595,12 @@ bool wait_for_input(int fd, bool trap, int timeout)
     assert(fd >= 0);
     if (fd >= FD_SETSIZE) {
 	xerror(0, Ngt("too many files are opened for yash to handle"));
-	return false;
+	return W_ERROR;
     }
 
+    if (trap)
+	sigint_received = false;
+
     ss = accept_sigmask;
     sigdelset(&ss, SIGCHLD);
     if (interactive_handlers_set) {
@@ -606,19 +625,28 @@ bool wait_for_input(int fd, bool trap, int timeout)
 #if YASH_ENABLE_LINEEDIT && defined SIGWINCH
 	handle_sigwinch();
 #endif
+	if (trap && sigint_received) {
+	    sigint_received = false;
+	    return W_INTERRUPTED;
+	}
 
 	fd_set fdset;
 	FD_ZERO(&fdset);
 	FD_SET(fd, &fdset);
-	if (pselect(fd + 1, &fdset, NULL, NULL, top, &ss) >= 0) {
-	    if (trap)
-		sigint_received = false;
-	    return FD_ISSET(fd, &fdset);
-	} else {
-	    if (errno != EINTR) {
-		xerror(errno, "pselect");
-		return false;
-	    }
+
+	int count = pselect(fd + 1, &fdset, NULL, NULL, top, &ss);
+
+	if (trap && sigint_received) {
+	    sigint_received = false;
+	    return W_INTERRUPTED;
+	}
+
+	if (count >= 0)
+	    return FD_ISSET(fd, &fdset) ? W_READY : W_TIMED_OUT;
+
+	if (errno != EINTR) {
+	    xerror(errno, "pselect");
+	    return W_ERROR;
 	}
     }
 }
@@ -682,7 +710,6 @@ int handle_traps(void)
 
     int signum = 0;
     bool save_sigint_received = sigint_received;
-    struct execstate_T *execstate = NULL;
     savelaststatus = laststatus;
 
     sigint_received = false;
@@ -696,16 +723,16 @@ int handle_traps(void)
 	    if (signal_received[i]) {
 		signal_received[i] = false;
 		wchar_t *command = trap_command[i];
-		if (command != NULL && command[0] != L'\0') {
+		if (!is_phantom && command != NULL && command[0] != L'\0') {
 #if YASH_ENABLE_LINEEDIT
 		    le_suspend_readline();
 #endif
-		    if (execstate == NULL)
-			execstate = save_execstate();
+		    struct execstate_T *execstate = save_execstate();
 		    signum = handled_signal = s->no;
 		    command = xwcsdup(command);
 		    exec_wcs(command, "trap", false);
 		    free(command);
+		    restore_execstate(execstate);
 		    laststatus = savelaststatus;
 		}
 	    }
@@ -718,16 +745,16 @@ int handle_traps(void)
 	    if (rtsignal_received[i]) {
 		rtsignal_received[i] = false;
 		wchar_t *command = rttrap_command[i];
-		if (command != NULL && command[0] != L'\0') {
+		if (!is_phantom && command != NULL && command[0] != L'\0') {
 #if YASH_ENABLE_LINEEDIT
 		    le_suspend_readline();
 #endif
-		    if (execstate == NULL)
-			execstate = save_execstate();
+		    struct execstate_T *execstate = save_execstate();
 		    signum = handled_signal = sigrtmin + i;
 		    command = xwcsdup(command);
 		    exec_wcs(command, "trap", false);
 		    free(command);
+		    restore_execstate(execstate);
 		    laststatus = savelaststatus;
 		}
 	    }
@@ -738,8 +765,6 @@ int handle_traps(void)
     sigint_received |= save_sigint_received;
     savelaststatus = -1;
     handled_signal = -1;
-    if (execstate != NULL)
-	restore_execstate(execstate);
 #if YASH_ENABLE_LINEEDIT
     if (shopt_notifyle && (le_state & LE_STATE_SUSPENDED))
 	print_job_status_all();
@@ -753,6 +778,9 @@ int handle_traps(void)
  * This function calls `reset_execstate' without saving `execstate'. */
 void execute_exit_trap(void)
 {
+    if (is_phantom)
+	return;
+
     wchar_t *command = trap_command[sigindex(0)];
     if (command != NULL) {
 #ifndef NDEBUG
@@ -773,6 +801,7 @@ void execute_exit_trap(void)
  * If `command' is NULL, the trap is reset to the default.
  * If `command' is an empty string, the trap is set to SIG_IGN.
  * This function may call `get_signal_name'.
+ * If `is_phantom` is true, `(rt)trap_command' is not modified.
  * An error message is printed to the standard error on error. */
 void set_trap(int signum, const wchar_t *command)
 {
@@ -813,15 +842,17 @@ void set_trap(int signum, const wchar_t *command)
 	return;
     }
 
-    free(*commandp);
-    if (command != NULL) {
-	if (command[0] != L'\0')
-	    any_trap_set = true;
-	*commandp = xwcsdup(command);
-    } else {
-	*commandp = NULL;
+    if (!is_phantom) {
+	free(*commandp);
+	if (command != NULL) {
+	    if (command[0] != L'\0')
+		any_trap_set = true;
+	    *commandp = xwcsdup(command);
+	} else {
+	    *commandp = NULL;
+	}
+	*receivedp = false;
     }
-    *receivedp = false;
     if (signum == 0)
 	return;
 
@@ -859,8 +890,10 @@ void set_trap(int signum, const wchar_t *command)
 	    if (interactive_handlers_set)
 		return;
 	    break;
+	case SIGTTIN:
+	case SIGTTOU:
 	case SIGTSTP:
-	    if (job_handler_set)
+	    if (job_handlers_set)
 		goto default_ignore;
 	    break;
 	case SIGTERM:
@@ -894,7 +927,8 @@ bool is_ignored(int signum)
     if (signum == 0)
 	return false;
 
-    if (doing_job_control_now && signum == SIGTSTP)
+    if (doing_job_control_now &&
+	    (signum == SIGTTIN || signum == SIGTTOU || signum == SIGTSTP))
 	return sigismember(&ignored_signals, signum);
 
     struct sigaction action;
@@ -909,12 +943,14 @@ void clear_exit_trap(void)
     set_trap(0, NULL);
 }
 
-/* Clears all traps except that are set to SIG_IGN. */
-void clear_traps(void)
+/* Change all traps into phantom except that are set to SIG_IGN. */
+void phantomize_traps(void)
 {
     if (!any_trap_set && !any_signal_received)
 	return;
 
+    is_phantom = true;
+
     {
 	size_t index = sigindex(0);
 	wchar_t *command = trap_command[index];
@@ -940,6 +976,40 @@ void clear_traps(void)
     any_trap_set = false, any_signal_received = false;
 }
 
+/* Remove all phantoms from `trap_command` and `rttrap_command`. */
+void banish_phantoms(void)
+{
+    if (!is_phantom)
+	return;
+    is_phantom = false;
+
+    {
+	size_t index = sigindex(0);
+	wchar_t *command = trap_command[index];
+	if (command != NULL && command[0] != L'\0') {
+	    free(command);
+	    trap_command[index] = NULL;
+	}
+    }
+    for (const signal_T *s = signals; s->no != 0; s++) {
+	size_t index = sigindex(s->no);
+	wchar_t *command = trap_command[index];
+	if (command != NULL && command[0] != L'\0') {
+	    free(command);
+	    trap_command[index] = NULL;
+	}
+    }
+#if defined SIGRTMIN && defined SIGRTMAX
+    for (int i = 0; i < RTSIZE; i++) {
+	wchar_t *command = rttrap_command[i];
+	if (command != NULL && command[0] != L'\0') {
+	    free(command);
+	    rttrap_command[i] = NULL;
+	}
+    }
+#endif
+}
+
 /* Tests the `sigint_received' flag. Returns true only if interactive. */
 bool is_interrupted(void)
 {
@@ -960,13 +1030,6 @@ void set_interrupted(void)
     sigint_received = true;
 }
 
-/* Resets the `sigint_received' flag. */
-/* This function should be called just after calling `handle_signals'. */
-void reset_sigint(void)
-{
-    sigint_received = false;
-}
-
 #if YASH_ENABLE_LINEEDIT
 
 #ifdef SIGWINCH
@@ -1162,6 +1225,7 @@ int trap_builtin(int argc, void **argv)
 	}
 
 	/* set traps */
+	banish_phantoms();
 	do {
 	    wchar_t *name = ARGV(xoptind);
 	    int signum = get_signal_number_toupper(name);
