@@ -1,6 +1,6 @@
 /* Yash: yet another shell */
 /* variable.c: deals with shell variables and parameters */
-/* (C) 2007-2015 magicant */
+/* (C) 2007-2016 magicant */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -230,6 +230,11 @@ void varkvfree(kvpair_T kv)
  * calls `varkvfree'. */
 void varkvfree_reexport(kvpair_T kv)
 {
+    if (kv.key == NULL) {
+	assert(kv.value == NULL);
+	return;
+    }
+
     variable_set(kv.key, NULL);
     if (((variable_T *) kv.value)->v_type & VF_EXPORT)
 	update_environment(kv.key);
@@ -734,10 +739,10 @@ bool do_assignments(const assign_T *assign, bool temp, bool export)
 
 	switch (assign->a_type) {
 	    case A_SCALAR:
-		value = expand_single(assign->a_scalar, TT_MULTI);
+		value = expand_single_and_unescape(
+			assign->a_scalar, TT_MULTI, true, false);
 		if (value == NULL)
 		    return false;
-		value = unescapefree(value);
 		if (shopt_xtrace)
 		    xtrace_variable(assign->a_name, value);
 		if (!set_variable(assign->a_name, value, scope, export))
@@ -1654,6 +1659,9 @@ int typeset_builtin(int argc, void **argv)
 	    || wcscmp(ARGV(0), L"set") == 0);
     }
 
+    if (function && global && wcscmp(ARGV(0), L"typeset") == 0)
+	return special_builtin_syntax_error(
+		mutually_exclusive_option_error(L'f', L'g'));
     if (function && export)
 	return special_builtin_syntax_error(
 		mutually_exclusive_option_error(L'f', L'x'));
@@ -1728,10 +1736,13 @@ int typeset_builtin(int argc, void **argv)
 		/* treat function */
 		function_T *f = ht_get(&functions, arg).value;
 		if (f != NULL) {
-		    if (readonly)
-			f->f_type |= VF_READONLY | VF_NODELETE;
-		    if (print)
-			print_function(arg, f, ARGV(0), readonly);
+		    if (print) {
+			if (!readonly || (f->f_type & VF_READONLY))
+			    print_function(arg, f, ARGV(0), readonly);
+		    } else {
+			if (readonly)
+			    f->f_type |= VF_READONLY | VF_NODELETE;
+		    }
 		} else {
 		    xerror(0, Ngt("no such function `%ls'"), arg);
 		}
@@ -2318,13 +2329,27 @@ const char unset_syntax[] = Ngt(
 );
 #endif
 
+/* Options for the "shift" built-in. */
+const struct xgetopt_T shift_options[] = {
+    { L'A', L"array", OPTARG_REQUIRED, false, NULL, },
+#if YASH_ENABLE_HELP
+    { L'-', L"help",  OPTARG_NONE,     false, NULL, },
+#endif
+    { L'\0', NULL, 0, false, NULL, },
+};
+
 /* The "shift" built-in */
 int shift_builtin(int argc, void **argv)
 {
+    const wchar_t *arrayname = NULL;
+
     const struct xgetopt_T *opt;
     xoptind = 0;
-    while ((opt = xgetopt(argv, help_option, 0)) != NULL) {
+    while ((opt = xgetopt(argv, shift_options, XGETOPT_DIGIT)) != NULL) {
 	switch (opt->shortopt) {
+	    case L'A':
+		arrayname = xoptarg;
+		break;
 #if YASH_ENABLE_HELP
 	    case L'-':
 		return print_builtin_help(ARGV(0));
@@ -2337,45 +2362,67 @@ int shift_builtin(int argc, void **argv)
     if (!validate_operand_count(argc - xoptind, 0, 1))
 	return special_builtin_syntax_error(Exit_ERROR);
 
-    size_t scount;
+    long count;
     if (xoptind < argc) {
-	long count;
 	if (!xwcstol(ARGV(xoptind), 10, &count)) {
 	    xerror(errno, Ngt("`%ls' is not a valid integer"), ARGV(xoptind));
 	    return special_builtin_syntax_error(Exit_ERROR);
-	} else if (count < 0) {
+	} else if (posixly_correct && count < 0) {
 	    xerror(0, Ngt("%ls: the operand value must not be negative"),
 		    ARGV(xoptind));
 	    return special_builtin_syntax_error(Exit_ERROR);
 	}
-#if LONG_MAX > SIZE_MAX
-	if (count > (long) SIZE_MAX)
-	    scount = SIZE_MAX;
-	else
-#endif
-	    scount = (size_t) count;
     } else {
-	scount = 1;
+	count = 1;
     }
 
-    variable_T *var = search_variable(L VAR_positional);
-    assert(var != NULL && (var->v_type & VF_MASK) == VF_ARRAY);
-    if (scount > var->v_valc) {
-	xerror(0,
-	    ngt("%zu: cannot shift so many "
-		"(there is only one positional parameter)",
-		"%zu: cannot shift so many "
-		"(there are only %zu positional parameters)",
-		var->v_valc),
-	    scount, var->v_valc);
+    variable_T *var;
+    if (arrayname == NULL) {
+	var = search_variable(L VAR_positional);
+	assert(var != NULL && (var->v_type & VF_MASK) == VF_ARRAY);
+    } else {
+	var = search_variable(arrayname);
+	if (wcschr(arrayname, L'=') != NULL ||
+		var == NULL || (var->v_type & VF_MASK) != VF_ARRAY) {
+	    xerror(0, Ngt("$%ls is not an array"), arrayname);
+	    return Exit_FAILURE;
+	}
+    }
+
+    unsigned long abscount =
+	(count >= 0) ? (unsigned long) count : -(unsigned long) count;
+
+    if (
+#if ULONG_MAX > SIZE_MAX
+	    abscount > (unsigned long) var->v_valc
+#else
+	    (size_t) abscount > var->v_valc
+#endif
+	    ) {
+	const char *message;
+	if (arrayname == NULL) {
+	    message = ngt("%ld: cannot shift so many "
+		    "(there is only one positional parameter)",
+		    "%ld: cannot shift so many "
+		    "(there are only %zu positional parameters)",
+		    var->v_valc);
+	} else {
+	    message = ngt("%ld: cannot shift so many "
+		    "(there is only one array element)",
+		    "%ld: cannot shift so many "
+		    "(there are only %zu array elements)",
+		    var->v_valc);
+	}
+	xerror(0, message, count, var->v_valc);
 	return Exit_FAILURE;
     }
 
+    size_t from = (count >= 0) ? 0 : (var->v_valc - (size_t) abscount);
     plist_T list;
     pl_initwith(&list, var->v_vals, var->v_valc);
-    for (size_t i = 0; i < scount; i++)
-	free(list.contents[i]);
-    pl_remove(&list, 0, scount);
+    for (size_t i = 0; i < (size_t) abscount; i++)
+	free(list.contents[from + i]);
+    pl_remove(&list, from, (size_t) abscount);
     var->v_valc = list.length;
     var->v_vals = pl_toary(&list);
 
@@ -2384,10 +2431,10 @@ int shift_builtin(int argc, void **argv)
 
 #if YASH_ENABLE_HELP
 const char shift_help[] = Ngt(
-"remove some positional parameters"
+"remove some positional parameters or array elements"
 );
 const char shift_syntax[] = Ngt(
-"\tshift [count]\n"
+"\tshift [-A array_name] [count]\n"
 );
 #endif
 
@@ -2669,6 +2716,7 @@ int read_builtin(int argc, void **argv)
     }
 
     /* split fields */
+    const wchar_t *tail;
     plist_T list;
     pl_init(&list);
     {
@@ -2676,7 +2724,7 @@ int read_builtin(int argc, void **argv)
 	if (ifs == NULL)
 	    ifs = DEFAULT_IFS;
 
-	extract_fields(buf.contents, split.contents, false, ifs, &list);
+	tail = extract_fields(buf.contents, split.contents, false, ifs, &list);
 	assert(list.length % 2 == 0);
     }
 
@@ -2702,8 +2750,11 @@ int read_builtin(int argc, void **argv)
     if (ro.array) {
 	assign_array(name, &list, 2 * (count - 1));
     } else {
-	const wchar_t *start = list.contents[2 * (count - 1)];
-	const wchar_t *end = list.contents[list.length - 1];
+	size_t i = count - 1;
+	const wchar_t *start = list.contents[2 * i];
+	const wchar_t *end = list.contents[2 * i + 1];
+	if (2 * count < list.length)
+	    end = tail;
 	wchar_t *field = xwcsndup(start, end - start);
 	set_variable(name, field, SCOPE_GLOBAL, shopt_allexport);
     }
@@ -3009,7 +3060,7 @@ variable_T *get_dirstack(void)
     variable_T *var = search_variable(L VAR_DIRSTACK);
     if (var != NULL) {
 	if ((var->v_type & VF_MASK) != VF_ARRAY) {
-	    xerror(0, Ngt("$DIRSTACK is not an array"));
+	    xerror(0, Ngt("$%ls is not an array"), L VAR_DIRSTACK);
 	    return NULL;
 	} else if (var->v_type & VF_READONLY) {
 	    xerror(0, Ngt("$%ls is read-only"), L VAR_DIRSTACK);
