@@ -19,15 +19,13 @@
 #include "common.h"
 #include "xfnmatch.h"
 #include <assert.h>
+#include <regex.h>
 #include <stdbool.h>
 #include <stdlib.h>
 #include <string.h>
 #include <wchar.h>
-#include <regex.h>
 #include "strbuf.h"
 #include "util.h"
-/* On old Mac OS X, <regex.h> requires `size_t' be already defined, hence it is
- * included after <stdlib.h>. */
 
 
 struct xfnmatch_T {
@@ -529,7 +527,10 @@ xfnmresult_T wmatch_literal(
 	    return MISMATCH;
 	if ((xfnm->flags & XFNM_TAILONLY) && (*s != L'\0'))
 	    return MISMATCH;
-	return (xfnmresult_T) { .start = 0, .end = xfnm->value.literal.length };
+	size_t slen = xfnm->value.literal.length;
+	if ((xfnm->flags & (XFNM_SHORTEST | XFNM_tailstar)) == XFNM_tailstar)
+	    slen += wcslen(s);
+	return (xfnmresult_T) { .start = 0, .end = slen };
     } else if (xfnm->flags & XFNM_TAILONLY) {
 	size_t slen = wcslen(s);
 	if (slen < xfnm->value.literal.length)
@@ -537,6 +538,8 @@ xfnmresult_T wmatch_literal(
 	size_t index = slen - xfnm->value.literal.length;
 	if (wcscmp(&s[index], xfnm->value.literal.contents) != 0)
 	    return MISMATCH;
+	if ((xfnm->flags & (XFNM_SHORTEST | XFNM_headstar)) == XFNM_headstar)
+	    index = 0;
 	return (xfnmresult_T) { .start = index, .end = slen };
     } else {
 	const wchar_t *ss;
@@ -730,4 +733,29 @@ void xfnm_free(xfnmatch_T *xfnm)
 }
 
 
+#if YASH_ENABLE_TEST
+
+/* Tests if extended regular expression `regex' matches string `s'. */
+bool match_regex(const wchar_t *s, const wchar_t *regex)
+{
+    regex_t compiled_regex;
+    char *mbs_regex = malloc_wcstombs(regex);
+    int err = regcomp(&compiled_regex, mbs_regex, REG_EXTENDED | REG_NOSUB);
+    free(mbs_regex);
+
+    if (err != 0)
+	return false;
+
+    char *mbs_s = malloc_wcstombs(s);
+    err = regexec(&compiled_regex, mbs_s, 0, NULL, 0);
+    free(mbs_s);
+
+    regfree(&compiled_regex);
+
+    return err == 0;
+}
+
+#endif /* YASH_ENABLE_TEST */
+
+
 /* vim: set ts=8 sts=4 sw=4 noet tw=80: */
